/*
 * kissdx - KiSS PC-Link Daemon eXtended (based on kissd)
 *
 * Copyright (C) 2005 Stelian Pop <stelian@popies.net>
 * Portions Copyright (C) 2006 Vidar Tysse <vidar@vidartysse.net>
 *
 * Heavily based on kiss4lin,
 * Copyright (C) 2004 Jacob Kolding <dacobi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "config.h"
#include "kissdx.h"
#include "utils.h"


// Global variables
config_settings_t config;	// Holds the current config settings as read from kissdx.conf


static int validate_media_path(const char *path, const char *persistentstoragepath)
{
	if (*persistentstoragepath && !strncmp(path, persistentstoragepath, strlen(persistentstoragepath))) {
		log("Media directory inside persistent storage directory is not allowed: %s", path);
		return 1;
	}
	return 0;
}

static int stat_outdated_path(const char *path, const char *subdir)
{
	char fullpath[PATH_MAX];
	struct stat statbuf;
	strcpy(fullpath, path);
	strncat(fullpath, "/", sizeof fullpath);
	strncat(fullpath, subdir, sizeof fullpath);
	if (stat(fullpath, &statbuf) == 0) {
		log("Old directory must be moved to new location: %s", fullpath);
		return 1;
	}
	return 0;
}

int parse_config(config_settings_t *config_p)
{
	FILE *f;
	char *home, *configpath;
	char buffer[512];
	int is_config_error = 0;

	// Set all config options to their default values
	memset(config_p, 0, sizeof (*config_p));
	strcpy(config_p->isofileextensions, "iso");
	strcpy(config_p->subtitle_catchall_pattern, "{name}*.*");
	config_p->max_recent_files = 30;
	config_p->picturemaxzoompercent = 100;
	config_p->picturecachetrimminginterval = 5;

	configpath = options.config_file;
	if (configpath != NULL) {
		if ((f = fopen(configpath, "r")) == NULL) {
			log("unable to open: %s", configpath);
			return 1;
		}
	}
	else {
		if ((home = getenv("HOME")) == NULL)
			home = "/";
		snprintf(buffer, sizeof(buffer), "%s/.kissdx.conf", home);
		configpath = buffer;
		if ((f = fopen(configpath, "r")) == NULL) {
			snprintf(buffer, sizeof(buffer), "%s/.kissd.conf", home);
			configpath = buffer;
			if ((f = fopen(configpath, "r")) == NULL) {
				configpath = "/etc/kissdx.conf";
				if ((f = fopen(configpath, "r")) == NULL) {
					configpath = "/etc/kissd.conf";
					if ((f = fopen(configpath, "r")) == NULL) {
						log0("config file not found!");
						log0("searched $HOME/ and /etc/ for kissdx.conf and kissd.conf");
						return 1;
					}
				}
			}
		}
	}

	if (options.verbose)
		log("Using config file %s", configpath);

	while (fgets(buffer, sizeof(buffer), f) != NULL) {
		char option[512], value[512];
		char *b = buffer;
		int rel_path = 0;

		for(; *b == ' ' || *b == '\t' || *b == '\n'; b++);
		if (b[0] == '#')
			continue;
		if (strlen(b) == 0)
			continue;

		if (sscanf(b, " %s = %[^\n]\n", option, value) != 2) {
			log("invalid line in config file: %s", buffer);
			is_config_error = 1;
		}

		if (!strcasecmp(option, "audiopath")) {
			strncpy(config_p->audiopath, value, sizeof(config_p->audiopath));
			config_p->audiopath[sizeof(config_p->audiopath) - 1] = '\0';
			rel_path = (value[0] != '/');
		}
		else if (!strcasecmp(option, "videopath")) {
			strncpy(config_p->videopath, value, sizeof(config_p->videopath));
			config_p->videopath[sizeof(config_p->videopath) - 1] = '\0';
			rel_path = (value[0] != '/');
		}
		else if (!strcasecmp(option, "picturepath")) {
			strncpy(config_p->picturepath, value, sizeof(config_p->picturepath));
			config_p->picturepath[sizeof(config_p->picturepath) - 1] = '\0';
			rel_path = (value[0] != '/');
		}
		else if (!strcasecmp(option, "persistentstoragepath")) {
			strncpy(config_p->persistentstoragepath, value, sizeof(config_p->persistentstoragepath));
			config_p->persistentstoragepath[sizeof(config_p->persistentstoragepath) - 1] = '\0';
			rel_path = (value[0] != '/');
		}
		else if (!strcasecmp(option, "kmlurl")) {
			strncpy(config_p->kmlurl, value, sizeof(config_p->kmlurl));
			config_p->kmlurl[sizeof(config_p->kmlurl) - 1] = '\0';
		}
		else if (!strcasecmp(option, "pretrigger")) {
			strncpy(config_p->pretrigger, value, sizeof(config_p->pretrigger));
			config_p->pretrigger[sizeof(config_p->pretrigger) - 1] = '\0';
			rel_path = (value[0] != '/');
		}
		else if (!strcasecmp(option, "posttrigger")) {
			strncpy(config_p->posttrigger, value, sizeof(config_p->posttrigger));
			config_p->posttrigger[sizeof(config_p->posttrigger) - 1] = '\0';
			rel_path = (value[0] != '/');
		}
		else if (!strcasecmp(option, "audiofileextensions")) {
			strncpy(config_p->audiofileextensions, value, sizeof(config_p->audiofileextensions));
			config_p->audiofileextensions[sizeof(config_p->audiofileextensions) - 1] = '\0';
			make_lowercase(config_p->audiofileextensions);
		}
		else if (!strcasecmp(option, "videofileextensions")) {
			strncpy(config_p->videofileextensions, value, sizeof(config_p->videofileextensions));
			config_p->videofileextensions[sizeof(config_p->videofileextensions) - 1] = '\0';
			make_lowercase(config_p->videofileextensions);
		}
		else if (!strcasecmp(option, "picturefileextensions")) {
			strncpy(config_p->picturefileextensions, value, sizeof(config_p->picturefileextensions));
			config_p->picturefileextensions[sizeof(config_p->picturefileextensions) - 1] = '\0';
			make_lowercase(config_p->picturefileextensions);
		}
		else if (!strcasecmp(option, "isofileextensions")) {
			strncpy(config_p->isofileextensions, value, sizeof(config_p->isofileextensions));
			config_p->isofileextensions[sizeof(config_p->isofileextensions) - 1] = '\0';
			make_lowercase(config_p->isofileextensions);
		}
		else if (!strcasecmp(option, "max_recent_files")) {
			config_p->max_recent_files = abs(atoi(value));
		}
		else if (!strcasecmp(option, "server_character_set")) {
			strncpy(config_p->server_character_set, value, sizeof(config_p->server_character_set));
			config_p->server_character_set[sizeof(config_p->server_character_set) - 1] = '\0';
		}
		else if (!strcasecmp(option, "client_character_set")) {
			strncpy(config_p->client_character_set, value, sizeof(config_p->client_character_set));
			config_p->client_character_set[sizeof(config_p->client_character_set) - 1] = '\0';
		}
		else if (!strcasecmp(option, "subtitlefilemapping")) {
			if (!strncmp(value, "*:", 2) && strlen(value) > 2) {
				strncpy(config_p->subtitle_catchall_pattern, value + 2, sizeof(config_p->subtitle_catchall_pattern));
				config_p->subtitle_catchall_pattern[sizeof(config_p->subtitle_catchall_pattern) - 1] = '\0';
			} else {
				log("Error in %s option in config file: %s", option, "Must start with '*:' followed by a filename pattern.");
				is_config_error = 1;
			}
		}
		else if (!strcasecmp(option, "picturetargetwidth")) {
			config_p->picturetargetwidth = abs(atoi(value));
		}
		else if (!strcasecmp(option, "picturetargetheight")) {
			config_p->picturetargetheight = abs(atoi(value));
		}
		else if (!strcasecmp(option, "picturemaxzoompercent")) {
			config_p->picturemaxzoompercent = abs(atoi(value));
		}
		else if (!strcasecmp(option, "picturecachesize")) {
			config_p->picturecachesize = atoi(value);
		}
		else if (!strcasecmp(option, "picturecachetrimminginterval")) {
			config_p->picturecachetrimminginterval = abs(atoi(value));
		}
		else if (!strcasecmp(option, "networktimeoutinterval")) {
			config_p->networktimeoutinterval = max(atoi(value), 0);
		}
		else if (!strcasecmp(option, "enablehiddenfilestext")) {
			if (!strchr(value, '/') && !strchr(value, '\\')) {
				strncpy(config_p->enablehiddenfilestext, value, sizeof(config_p->enablehiddenfilestext));
				config_p->enablehiddenfilestext[sizeof(config_p->enablehiddenfilestext) - 1] = '\0';
			} else {
				log("Error in %s option in config file: %s", option, "Can not contain file system delimiters '/' and '\\'.");
				is_config_error = 1;
			}
		}
		else if (!strcasecmp(option, "enablehiddenfilesminutes")) {
			config_p->enablehiddenfilesminutes = max(atoi(value), 0);
		}
		else {
			log("unknown option in config file: %s", option);
			is_config_error = 1;
		}

		if (rel_path) {
			log("paths must be absolute refs in config file: %s", value);
			is_config_error = 1;
		}
	}

	if (config_p->audiopath[0] == '\0') {
		log0("audiopath not set in config file");
		is_config_error = 1;
	}

	if (config_p->videopath[0] == '\0') {
		log0("videopath not set in config file");
		is_config_error = 1;
	}

	if (config_p->picturepath[0] == '\0') {
		log0("picturepath not set in config file");
		is_config_error = 1;
	}

	clean_pathname(config_p->audiopath);
	clean_pathname(config_p->videopath);
	clean_pathname(config_p->picturepath);

	if (config_p->persistentstoragepath[0] != '\0' && !is_config_error) {
		clean_pathname(config_p->persistentstoragepath);

		// Check whether any media directories are inside the persistent storage directory.  If they are, it's a config error (we risk name collisions for .recent/.piccache).
		if (validate_media_path(config_p->audiopath, config_p->persistentstoragepath))
			is_config_error = 1;
		if (validate_media_path(config_p->videopath, config_p->persistentstoragepath))
			is_config_error = 1;
		if (validate_media_path(config_p->picturepath, config_p->persistentstoragepath))
			is_config_error = 1;

		if (!is_config_error) {
			// Look for .recent directories inside media directories.  If they are still there, it's a config error.
			int is_recent_there = 0;
			if (stat_outdated_path(config_p->audiopath, ".recent"))
				is_recent_there = 1;
			if (stat_outdated_path(config_p->videopath, ".recent"))
				is_recent_there = 1;
			if (stat_outdated_path(config_p->picturepath, ".recent"))
				is_recent_there = 1;
			if (is_recent_there) {
				log("The .recent directories must be deleted or moved to the new subdirectories of %s", config_p->persistentstoragepath);
				is_config_error = 1;
			}
			if (stat_outdated_path(config_p->picturepath, ".piccache")) {
				log("The .piccache directory must be deleted or moved to %s", config_p->persistentstoragepath);
				is_config_error = 1;
			}
	
			// Create the .recent directories now (whether error or not) to aid user in moving stuff (and to aid ourselves later...)
			char path[PATH_MAX];
			if (mkdir(config_p->persistentstoragepath, standard_dir_filemode) >= 0)
				chmod(config_p->persistentstoragepath, standard_dir_filemode);
	
			strcpy(path, config_p->persistentstoragepath);
			strncat(path, "/audio", sizeof path);
			if (mkdir(path, standard_dir_filemode) >= 0)
				chmod(path, standard_dir_filemode);
			strncat(path, "/.recent", sizeof path);
			if (mkdir(path, standard_dir_filemode) >= 0)
				chmod(path, standard_dir_filemode);
	
			strcpy(path, config_p->persistentstoragepath);
			strncat(path, "/video", sizeof path);
			if (mkdir(path, standard_dir_filemode) >= 0)
				chmod(path, standard_dir_filemode);
			strncat(path, "/.recent", sizeof path);
			if (mkdir(path, standard_dir_filemode) >= 0)
				chmod(path, standard_dir_filemode);
	
			strcpy(path, config_p->persistentstoragepath);
			strncat(path, "/picture", sizeof path);
			if (mkdir(path, standard_dir_filemode) >= 0)
				chmod(path, standard_dir_filemode);
			strncat(path, "/.recent", sizeof path);
			if (mkdir(path, standard_dir_filemode) >= 0)
				chmod(path, standard_dir_filemode);
		}

	}

	if (options.kml_forwarder && config_p->kmlurl[0] == '\0') {
		log0("kmlurl not set in config file");
		is_config_error = 1;
	}

	fclose(f);

	if (is_config_error) {
		log0("Error(s) in the configuration file prevent kissdx from using the file.");
		return 1;
	}

	show_current_config_Verbose(config_p);
	return 0;
}

void show_current_config_Verbose(const config_settings_t *config_p)
{
	if (options.verbose) {
		log("kissdx %s is now using the following configuration:", KISSD_VERSION);
		log("audiopath = %s",						config_p->audiopath);
		log("videopath = %s",						config_p->videopath);
		log("picturepath = %s",						config_p->picturepath);
		log("persistentstoragepath = %s",			config_p->persistentstoragepath);
		log("kmlurl = %s",							config_p->kmlurl);
		log("pretrigger = %s",						config_p->pretrigger);
		log("posttrigger = %s", 					config_p->posttrigger);
		log("audiofileextensions = %s", 			config_p->audiofileextensions);
		log("videofileextensions = %s",				config_p->videofileextensions);
		log("picturefileextensions = %s",			config_p->picturefileextensions);
		log("isofileextensions = %s",				config_p->isofileextensions);
		log("max_recent_files = %d",				config_p->max_recent_files);
		log("server_character_set = %s",			config_p->server_character_set);
		log("client_character_set = %s",			config_p->client_character_set);
		log("subtitle_catchall_pattern = %s",		config_p->subtitle_catchall_pattern);
		log("picturetargetwidth = %d",				config_p->picturetargetwidth);
		log("picturetargetheight = %d",				config_p->picturetargetheight);
		log("picturemaxzoompercent = %d",			config_p->picturemaxzoompercent);
		log("picturecachesize = %d",				config_p->picturecachesize);
		log("picturecachetrimminginterval = %d",	config_p->picturecachetrimminginterval);
		log("networktimeoutinterval = %d",			config_p->networktimeoutinterval);
		log("enablehiddenfilestext = %s",			config_p->enablehiddenfilestext);
		log("enablehiddenfilesminutes = %d",		config_p->enablehiddenfilesminutes);
	}
}
